#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <selinux/selinux.h>
#include <selinux/context.h>

#include "altselinux.h"

#define ALTSELINUX_PATH_SEAPPLET	"/usr/bin/seapplet"
#define ALTSELINUX_PATH_LAUNCHER	"/usr/bin/slrun"

static int _altselinux_status = -1;

static int _altselinux_check()
{
	struct stat stat_file;

	if (is_selinux_mls_enabled() <= 0)
	{
		fprintf(stderr, "altselinux: check: selinux is disabled or not mls\n");
		return 1;
	}

	if (stat(ALTSELINUX_PATH_SEAPPLET, &stat_file) != 0 || !(stat_file.st_mode & S_IXUSR))
	{
		fprintf(stderr, "altselinux: check: seapplet (%s) not found\n", ALTSELINUX_PATH_SEAPPLET);
		return 2;
	}
	if (stat(ALTSELINUX_PATH_LAUNCHER, &stat_file) != 0 || !(stat_file.st_mode & S_IXUSR))
	{
		fprintf(stderr, "altselinux: check: launcher (%s) not found\n", ALTSELINUX_PATH_LAUNCHER);
		return 3;
	}
	return 0;
}

int altselinux_is_enabled()
{
	if (_altselinux_status > 0) return 0;
	else if (_altselinux_status == 0) return 1;
	_altselinux_status = _altselinux_check();
	return altselinux_is_enabled();
}

static char* _exec_write_read(char *cmd, char **arg, char *in, int in_size)
{
#define UTIL_BUF_SIZE	1024
	int pipe_in[2];
	int pipe_out[2];

	if (pipe(pipe_in) == -1) return NULL;
	if (pipe(pipe_out) == -1)
	{
		close(pipe_in[0]);
		close(pipe_in[1]);
		return NULL;
	}

	pid_t pid = fork();
	if (pid > 0)
	{
		char *result;
		char buf[UTIL_BUF_SIZE];
		int len, size;
		int status;

		close(pipe_in[0]);
		close(pipe_out[1]);

		result = (char*)malloc(1);

		if (in != NULL)
		{
			if (write(pipe_in[1], in, in_size) == -1)
			{
				close(pipe_in[1]);
				close(pipe_out[0]);
				free(result);
				return NULL;
			}
		}
		close(pipe_in[1]);

		size = 0;
		while ((len = read(pipe_out[0], buf, UTIL_BUF_SIZE)) > 0)
		{
			result = (char*)realloc(result, size + len + 1);
			if (result == NULL)
			{
				close(pipe_out[0]);
				return NULL;
			}
			memcpy(result + size, buf, len);
			size += len;
		}
		close(pipe_out[0]);

		if (wait(&status) == -1 || (!WIFEXITED(status) || (WEXITSTATUS(status) != 0)))
		{
			free(result);
			return NULL;
		}

		result[size] = 0;

		return result;
	}
	else if (pid == 0)
	{
		dup2(pipe_in[0], STDIN_FILENO);
		dup2(pipe_out[1], STDOUT_FILENO);

		close(pipe_in[0]);
		close(pipe_out[1]);

		execv(cmd, arg);

		_exit(1);
	}

	return NULL;
}

static char* _get_context_color(char *context)
{
	char *argv[4];

	argv[0] = ALTSELINUX_PATH_SEAPPLET;
	argv[1] = "color";
	argv[2] = context;
	argv[3] = NULL;

	return _exec_write_read(argv[0], argv, NULL, 0);
}

int altselinux_get_context_color(char* context, unsigned char* r, unsigned char* g, unsigned char* b)
{
	char *str_color;

	if (context == NULL || r == NULL || g == NULL || b == NULL)
	{
		fprintf(stderr, "altselinux: get_context_color: bad value\n");
		return 1;
	}

	str_color = _get_context_color(context);
	if (str_color == NULL)
	{
		fprintf(stderr, "altselinux: get_context_color: error exec seapplet\n");
		return 2;
	}

	if (sscanf(str_color, "%hhu %hhu %hhu", r, g, b) != 3)
	{
		fprintf(stderr, "altselinux: get_context_color: seapplet return bad string\n");
		free(str_color);
		return 3;
	}

	free(str_color);
	return 0;
}

char* altselinux_get_new_level()
{
	char *argv[3];
	char *result;

	argv[0] = ALTSELINUX_PATH_SEAPPLET;
	argv[1] = "print";
	argv[2] = NULL;

	result = _exec_write_read(argv[0], argv, NULL, 0);
	if (result == NULL) fprintf(stderr, "altselinux: get_new_level: error exec seapplet\n");

	return result;
}

int altselinux_need_launcher(char **new_level_out)
{
	char *new_level;
	char *cur_level;

	new_level = altselinux_get_new_level();
	if (new_level == NULL) return 0;

	cur_level = altselinux_get_current_level();
	if (cur_level == NULL)
	{
		free(new_level);
		return 0;
	}

	if (strlen(new_level) > 1 && strlen(cur_level) > 1)
	{
		int new_level_i;
		int cur_level_i;

		new_level_i = atoi(new_level + 1);
		cur_level_i = atoi(cur_level + 1);
		if (new_level_i >= cur_level_i && strcmp(new_level, cur_level) != 0)
		{
			if (new_level_out != NULL) *new_level_out = new_level;
			else free(new_level);

			free(cur_level);
			return 1;
		}
	}

	free(cur_level);
	free(new_level);
	return 0;
}

static char* _get_lowlevel_from_range(const char *context_range)
{
	char *rlevel;
	char *range;
	char *ch;

	if (context_range == NULL) return NULL;

	range = strdup(context_range);

	ch = strchr(range, '-');
	if (ch != NULL) *ch = 0;

	rlevel = strdup(range);

	free(range);
	return rlevel;
}

char* altselinux_get_current_level()
{
	char *result;
	char *con;
	const char *context_range;
	context_t context;

	if (getcon_raw(&con) < 0)
	{
		fprintf(stderr, "altselinux: get_current_level: getcon_raw() < 0\n");
		return NULL;
	}

	context = context_new(con);
	context_range = context_range_get(context);

	result = _get_lowlevel_from_range(context_range);

	context_free(context);
	freecon(con);
	return result;
}

char* altselinux_get_path(unsigned int id)
{
	if (id == ALTSELINUX_ID_SEAPPLET) return ALTSELINUX_PATH_SEAPPLET;
	else if (id == ALTSELINUX_ID_LAUNCHER) return ALTSELINUX_PATH_LAUNCHER;
	return NULL;
}
